# Generated by Django 5.2.5 on 2025-09-29 14:58
from __future__ import annotations

from django.db import migrations, models

import schematools.contrib.django.validators


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0022_inline_tables_for_all_versions_in_schema_data"),
    ]

    operations = [
        migrations.CreateModel(
            name="Publisher",
            fields=[
                ("id", models.CharField(max_length=100, primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=100)),
                (
                    "schema_data",
                    models.TextField(
                        validators=[schematools.contrib.django.validators.validate_json],
                        verbose_name="Amsterdam Schema Contents",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Scope",
            fields=[
                ("id", models.CharField(max_length=100, primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=100)),
                (
                    "schema_data",
                    models.TextField(
                        validators=[schematools.contrib.django.validators.validate_json],
                        verbose_name="Amsterdam Schema Contents",
                    ),
                ),
            ],
        ),
    ]
