# Generated by Django 5.1.4 on 2025-12-02 10:48
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0024_alter_profile_id"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="datasetversion",
            name="lifecycle_status",
        ),
        migrations.AddField(
            model_name="datasetversion",
            name="status",
            field=models.CharField(
                choices=[("D", "under_development"), ("S", "stable"), ("U", "superseded")],
                default="D",
            ),
        ),
    ]
